# Copyright 2023-2023 by AccidentallyTheCable <cableninja@cableninja.net>.
# All rights reserved.
# This file is part of AccidentallyTheCables Utility Kit,
# and is released under "GPLv3". Please see the LICENSE
# file that should have been included as part of this package.
#### END COPYRIGHT BLOCK ###



# Copyright 2023-2023 by AccidentallyTheCable <cableninja@cableninja.net>.
# All rights reserved.
# This file is part of AccidentallyTheCables Utility Kit,
# and is released under "GPLv3". Please see the LICENSE
# file that should have been included as part of this package.
#### END COPYRIGHT BLOCK ###
import re
import sys
import signal
import logging
import typing
from os import kill,getpid
from pathlib import Path
from json import loads as json_parse
from json import dumps as json_dump
from yaml import dump as yaml_dump
from yaml import safe_load as yaml_parse
from tomlkit import parse as toml_parse
from tomlkit import dumps as toml_dump

def toml_parser(content:str) -> dict[typing.Any,typing.Any]:
    """TOML Parse and unwrap
    @param str \c content Raw TOML string
    @retval dict[Any,Any] Dictionary from TOML
    """
    return toml_parse(content).unwrap()

class UtilFuncs:
    """Utility Functions
    """

    FILE_TYPE_MAP:dict[str,typing.Callable] = {
        ".yml": yaml_parse,
        ".yaml": yaml_parse,
        ".json": json_parse,
        ".toml": toml_parser
    }
    TYPE_OVERRIDE_MAP:dict[str,typing.Callable] = {
        "yaml": yaml_parse,
        "json": json_parse,
        "toml": toml_parser
    }
    OUTPUT_TYPE_MAP:dict[str,typing.Callable] = {
        "yaml": yaml_dump,
        "json": json_dump,
        "toml": toml_dump
    }

    CONFIG_SCAN_BASE_PATHS:list[Path] = [
        Path("~/.local/"),
        Path("/etc/")
    ]

    CONFIG_SCAN_FILE_EXTS:list[str] = [
        "json",
        "yaml",
        "toml"
    ]

    shutdown:bool = False
    restart:bool = False

    @staticmethod
    def create_object_logger(obj:object) -> logging.Logger:
        """Create logging.Logger with Specified object Class Name
        @param object \c obj Object to Create Logger for
        @retval logging.Logger Logger Instance
        """
        classname:str = str(obj.__class__.__qualname__)
        return logging.getLogger(classname)

    @staticmethod
    def create_static_logger(classname:str) -> logging.Logger:
        """Create logging.Logger with Specified Name
        @param str \c classname Name of Logger to Create
        @retval logging.Logger Logger Instance
        """
        return logging.getLogger(classname)

    @staticmethod
    def dump_sstr(content:dict[typing.Any,typing.Any],output_type:str,**kwargs:typing.Any) -> str:
        """Dump Structured Data to String, Selectable Output Type.
        @param dict[Any,Any] \c content Content to dump
        @param str \c output_type Output Type, Allowed: yaml, json, toml
        @param kwargs \c kwargs kwargs passed to selected data dumper
        @retval str Structured Data as a string
        @exception LookupError Raised When Unable to Locate a a parser by file type, or the overridden type is invalid
        @exception RuntimeError Raised When Unable to Parse the specified file using the specified type (automatically, or overridden)

        kwargs:
            See the related kwargs for:
                json.dumps
                yaml.dump
                tomlkit.dumps
        """
        if output_type not in UtilFuncs.OUTPUT_TYPE_MAP.keys():
            raise LookupError("Output Type is not valid",output_type)
        dumper:typing.Callable = UtilFuncs.OUTPUT_TYPE_MAP[output_type]
        try:
            output:str = dumper(content,*kwargs)
        except BaseException as e:
            raise RuntimeError("Failed to output content",'.'.join([dumper.__class__.__name__,dumper.__name__]),e) from e
        return output

    @staticmethod
    def load_sstr(raw_data:str,content_type:str) -> dict[typing.Any,typing.Any]:
        """Load Structured Data Content, based on content_type
        @param str \c raw_data Raw Content
        @param str \c content_type Render for Content Type, Allowed: yaml, json, toml
        @retval dict[Any,Any] Dictionary of Structured Data
        """
        parser:typing.Callable
        if content_type not in UtilFuncs.TYPE_OVERRIDE_MAP.keys():
            raise LookupError("Override Type is not valid",content_type)
        parser = UtilFuncs.TYPE_OVERRIDE_MAP[content_type]
        loaded_config:dict[typing.Any,typing.Any] = parser(raw_data)
        return loaded_config

    @staticmethod
    def load_sfile(target_file:Path,override_type:str = "auto") -> dict[typing.Any,typing.Any]:
        """Load Structured Data File, automatically determining data by file extension
        @param Path \c target_file Configuration File to Load
        @param str \c override_type Input Type Override, Allowed: yaml, json, toml, auto. Default: auto
        @retval dict[Any,Any] Dictionary of Structured Data
        @exception LookupError Raised When Unable to Locate a a parser by file type, or the overridden type is invalid
        @exception RuntimeError Raised When Unable to Parse the specified file using the specified type (automatically, or overridden)

        Accepted File Extensions:
            YAML: .yaml, .yml
            JSON: .json
            TOML: .toml
        """
        if not target_file.exists():
            raise FileNotFoundError("Cannot Locate File",target_file.as_posix())
        content_type:str = override_type
        if override_type == "auto":
            if target_file.suffix not in UtilFuncs.FILE_TYPE_MAP.keys():
                raise LookupError("No Parser located for file type",target_file.suffix)
            content_type = re.sub(r'^\.','',target_file.suffix)
        try:
            with open(target_file,"r",encoding="utf-8") as f:
                loaded_config:dict[typing.Any,typing.Any] = UtilFuncs.load_sstr(f.read(),content_type)
                return loaded_config
        except BaseException as e:
            raise RuntimeError("Failed to Parse file",target_file.as_posix(),e) from e

    @staticmethod
    def find_config_file(service:str,config_name:str) -> typing.Union[Path,None]:
        """Search for Configuration File in common places

        Uses UtilFuncs.CONFIG_SCAN_BASE_PATHS and UtilFuncs.CONFIG_SCAN_FILE_EXTS to search for files

        Scans as: <scan path>/<service>/<config_name>.<scan ext>

        @param str \c service Name of Service to Use (EX: myservice)
        @param str \c config_name Name of Configration File to find
        @retval Union[Path,None] Located configuration path, or None if not found
        """
        for p in UtilFuncs.CONFIG_SCAN_BASE_PATHS:
            check_svcpath:Path = p.joinpath(service).resolve()
            if not check_svcpath.is_dir():
                continue
            for ext in UtilFuncs.CONFIG_SCAN_FILE_EXTS:
                check_file:Path = check_svcpath.joinpath(f"{config_name}.{ext}").resolve()
                if check_file.is_file():
                    return check_file
        return None

    @staticmethod
    def add_config_search_path(path:Path) -> None:
        """Add Basepath to Config Search (EX /usr/local/)
        @param Path \c path Basepath to add for scanning
        @retval None Nothing
        """
        if path not in UtilFuncs.CONFIG_SCAN_BASE_PATHS:
            UtilFuncs.CONFIG_SCAN_BASE_PATHS.append(path)

    @staticmethod
    def remove_config_search_path(path:Path) -> None:
        """Remove Basepath from Config Search
        @param Path \c path Basepath to attempt to remove
        @retval None Nothing
        """
        try:
            UtilFuncs.CONFIG_SCAN_BASE_PATHS.pop(UtilFuncs.CONFIG_SCAN_BASE_PATHS.index(path))
        except ValueError:
            return

    @staticmethod
    def add_config_search_file_ext(ext:str) -> None:
        """Add File Extension to Config Search (EX conf)
        @param str \c ext File Extension to add for scanning
        @retval None Nothing
        """
        if ext not in UtilFuncs.CONFIG_SCAN_FILE_EXTS:
            UtilFuncs.CONFIG_SCAN_FILE_EXTS.append(ext)

    @staticmethod
    def remove_config_search_file_ext(ext:str) -> None:
        """Remove File Extension from Config Search
        @param str \c ext File Extension to attempt to remove
        @retval None Nothing
        """
        try:
            UtilFuncs.CONFIG_SCAN_FILE_EXTS.pop(UtilFuncs.CONFIG_SCAN_FILE_EXTS.index(ext))
        except ValueError:
            return

    @staticmethod
    def scan_dir(target_path:Path,
                 callback:typing.Callable[[Path,dict[str,typing.Any]],None],
                 callback_data:dict[str,typing.Any],
                 exclude_dirs:typing.Optional[list[re.Pattern]] = None,
                 exclude_files:typing.Optional[list[re.Pattern]] = None,
                 include_files:typing.Optional[list[re.Pattern]] = None
                ) -> None:
        """Scan A Directory, and Execute callback on discovered files, that do not match the exclusions
        @param Path \c target_path Path to Scan for Files
        @param typing.Callable[[Path,dict[str,Any]],None] \c callback Callback function to execute on each file
        @param dict[str,Any] \c callback_data Data to pass to the callback function
        @param list[re.Pattern] \c exclude_dirs (optional) Regex Compiled list of directory patterns to exclude
        @param list[re.Pattern] \c exclude_files (optional) Regex Compiled list of file patterns to exclude
        @param list[re.Pattern] \c include_files (optional) Regex Compiled list of file patterns to include
        """
        files:typing.Generator[Path, None, None] = target_path.glob("*")
        skip:bool = False
        for file in files:
            file_path:Path = Path(file)
            if file_path.is_dir():
                skip = False
                if exclude_dirs is not None:
                    for reg in exclude_dirs:
                        if reg.match(file_path.name):
                            skip = True
                            break
                if not skip:
                    UtilFuncs.scan_dir(target_path=file_path,callback=callback,callback_data=callback_data,exclude_dirs=exclude_dirs,exclude_files=exclude_files)
            if file_path.is_file():
                if include_files is not None:
                    skip = True
                    for reg in include_files:
                        if reg.match(file_path.name):
                            skip = False
                            break
                if exclude_files is not None:
                    skip = False
                    for reg in exclude_files:
                        if reg.match(file_path.name):
                            skip = True
                            break
                if not skip:
                    callback(file_path,callback_data)

    @staticmethod
    def deep_sort(input:dict[str,typing.Any]) -> dict[str,typing.Any]:
        """Deep Sort Dictionaries of varying data
        @param dict[str,typing.Any] \c input Input Dictionary
        @retval dict[str,typing.Any] New Sorted Dictionary
        """
        new_dict:dict[str,typing.Any] = {}
        for k,v in input.items():
            if isinstance(v,dict):
                new_dict[k] = dict(sorted(v.items()))
            elif isinstance(v,list):
                new_list:list[typing.Any] = []
                for i in v:
                    if isinstance(i,dict):
                        new_list.append(UtilFuncs.deep_sort(i))
                    else:
                        new_list.append(i)
                new_dict[k] = new_list
            else:
                new_dict[k] = v
        return new_dict

    @staticmethod
    def register_pid(service:str,pid_dir:Path = Path("/var/run/")) -> None:
        """Write Process ID to pid file at <pid_dir>/<service>.pid
        @param str service Service Name
        @param Path pid_dir Directory to write pid file to
        @retval None Nothing
        """
        pid_file:Path = pid_dir.joinpath(f"{service}.pid")
        if not pid_dir.exists():
            pid_dir.mkdir(parents=True,exist_ok=True)
        if pid_file.exists():
            with open(pid_file, "r", encoding="utf-8") as f:
                read_pid:int = int(f.read())
            if UtilFuncs.check_pid(read_pid):
                raise SystemError("Cannot start another process, one is already running",read_pid)
        with open(pid_file, "w", encoding="utf-8") as f:
            f.write(str(getpid()))

    @staticmethod
    def check_pid(pid:int) -> bool:
        """Check if PID exists (via os.kill(..,0))
        @param int \c pid PID to check
        @retval bool Whether PID exists or not
        """
        try:
            kill(pid,0)
        except OSError:
            return False
        return True

    # pylint: disable=unused-argument
    @staticmethod
    def sighandler(signum:int, frame:typing.Any) -> None:
        """Signal Handler
        @param signal.Signals \c signum Raised Signal
        @param Any \c frame Frame which raised the signal
        @retval None Nothing
        """
        logging.warning("Signal thrown")

        restart_signals:list[signal.Signals] = []
        shut_signals:list[signal.Signals] = []
        if sys.platform == "win32":
            shut_signals = [ signal.SIGINT, signal.CTRL_C_EVENT ]
            restart_signals = [ ]
        else:
            shut_signals = [ signal.SIGABRT, signal.SIGILL, signal.SIGINT, signal.SIGTERM ]
            restart_signals = [ signal.SIGHUP ]

        if signum in shut_signals:
            UtilFuncs.shutdown = True
            logging.info("Shutting Down")
            return
        if signum in restart_signals:
            logging.info("Reloading Service")
            UtilFuncs.restart = True
            return
    # pylint: enable=unused-argument

    @staticmethod
    def register_signals() -> None:
        """Register Signal Handlers
        @retval None Nothing
        """
        signals:list[signal.Signals] = []
        if sys.platform == "win32":
            signals = [ signal.SIGINT ]
        else:
            signals = [ signal.SIGABRT, signal.SIGILL, signal.SIGINT, signal.SIGTERM, signal.SIGHUP ]
        for sig in signals:
            signal.signal(sig,UtilFuncs.sighandler)



# Copyright 2023-2023 by AccidentallyTheCable <cableninja@cableninja.net>.
# All rights reserved.
# This file is part of AccidentallyTheCables Utility Kit,
# and is released under "GPLv3". Please see the LICENSE
# file that should have been included as part of this package.
#### END COPYRIGHT BLOCK ###
from typing import Callable
import typing_extensions

class FunctionSubscriber:
    """ Function subscriber


    Provides a += / -= interface for adding callback functions

    Example:

    \li Define some functions

        \code{.py}
        def a():
            print("a")
        def b():
            print("b")
        \endcode

    \li Create a subscriber and Subscribe the above functions
        \code{.py}
        f = FunctionSubscriber()
        f += a
        f += b
        print(f.functions)
        \endcode

    \li Remove a subscription
        \code{.py}
        f -= a
        print(f.functions)
        \endcode

    \li Execute subscribed functions
        \code{.py}
        for method_def in f.functions:
            method_def()
        \endcode
    """

    _functions:list[Callable]

    @property
    def functions(self) -> list[Callable]:
        """\b \e PROPERTY; Currently Subscribed Functions"""
        return self._functions

    def __init__(self) -> None:
        """Initializer
        """
        self._functions = []

    def __iadd__(self,fn:Callable) -> typing_extensions.Self:
        """Inline Add. Subscribe Function
        @param method \c fn Method to Subscribe
        """
        if fn not in self._functions:
            self._functions.append(fn)
        return self

    def __isub__(self,fn:Callable) -> typing_extensions.Self:
        """Inline Subtract. Unsubscribe Function
        @param method \c fn Method to Unsubscribe
        """
        if fn not in self._functions:
            return self
        i:int = self._functions.index(fn)
        self._functions.pop(i)
        return self



# Copyright 2023-2023 by AccidentallyTheCable <cableninja@cableninja.net>.
# All rights reserved.
# This file is part of AccidentallyTheCables Utility Kit,
# and is released under "GPLv3". Please see the LICENSE
# file that should have been included as part of this package.
#### END COPYRIGHT BLOCK ###
import logging
import threading
import typing
from pathlib import Path
from time import sleep
from sys import exit as sys_exit

from atckit.subscriber import FunctionSubscriber
from atckit.utilfuncs import UtilFuncs

class Service:
    """Service Controller
    Modify Service threads by adding to or subtracting from Service.services (See FunctionSubscriber)
    """

    _SERVICE_NAME:str# = "myservice"
    _SERVICE_SHUTDOWN_LIMIT:int# = 300
    _SERVICE_CHECK_TIME:float# = 0.5
    services:FunctionSubscriber

    _config_file:Path
    _config_path:Path
    _pid_dir:Path
    _pid_file:Path
    _config:dict[str,typing.Any]
    _service_threads:dict[typing.Callable,threading.Thread]

    @property
    def should_run(self) -> bool:
        """Should be running Flag
        Combine shutdown and restart
        @retval bool Should be Running
        """
        return not (self.shutdown or self.restart)

    @property
    def shutdown(self) -> bool:
        """Shutdown Flag
        Passthrough to Utilfuncs.shutdown
        """
        return UtilFuncs.shutdown

    @shutdown.setter
    def shutdown(self,value:bool) -> None:
        """Shutdown Flag
        Passthrough to Utilfuncs.shutdown
        @param bool value Flag Value
        """
        UtilFuncs.shutdown = value

    @property
    def restart(self) -> bool:
        """Restart Flag
        Passthrough to Utilfuncs.restart
        """
        return UtilFuncs.restart

    @restart.setter
    def restart(self,value:bool) -> None:
        """Restart Flag
        Passthrough to Utilfuncs.restart
        @param bool value Flag Value
        """
        UtilFuncs.restart = value

    logger:logging.Logger

    def __init__(self) -> None:
        """Initializer
        Read Config
        Write PID
        @throws ValueError: invalid self._mode
        """
        if not hasattr(self,"_SERVICE_CHECK_TIME"):
            self._SERVICE_CHECK_TIME = 0.5
        if not hasattr(self,"_SERVICE_SHUTDOWN_LIMIT"):
            self._SERVICE_SHUTDOWN_LIMIT = 300
        if not hasattr(self,"_SERVICE_NAME"):
            raise ValueError(f"Service name must be defined in {type(self).__qualname__}._SERVICE_NAME")
        UtilFuncs.register_signals()
        self.logger = UtilFuncs.create_object_logger(self)
        self.services = FunctionSubscriber()
        self._service_threads = {}
        config_path:typing.Union[Path,None] = UtilFuncs.find_config_file(self._SERVICE_NAME,self._SERVICE_NAME)
        if config_path is None:
            self.logger.critical(f"Unable to locate configuration file for {self._SERVICE_NAME}")
            self._config = {}
        else:
            self._config_file = config_path
            self._config_path = config_path.parent
            config_file_str:str = self._config_file.as_posix()
            self.logger.debug(f"Loading {config_file_str}")
            self._config = UtilFuncs.load_sfile(self._config_file)

        if "pid_dir" not in self._config.keys():
            self._pid_dir = Path(f"/var/run/{self._SERVICE_NAME}/")
        else:
            self._pid_dir = Path(self._config["pid_dir"]).resolve()
        self._pid_file = self._pid_dir.joinpath(f"{self._SERVICE_NAME}.pid")
        try:
            UtilFuncs.register_pid(self._SERVICE_NAME,self._pid_dir)
        except BaseException as e:
            self.logger.critical("Register PID failed, See Exception information",exc_info=e)
            sys_exit(1)

    def run(self) -> None:
        """Start Service Thread(s)
        @retval None Nothing
        """
        while not self.shutdown:
            for f in self.services.functions:
                t:threading.Thread
                if f in self._service_threads.keys():
                    t = self._service_threads[f]
                    if not t.is_alive():
                        self.logger.warning(f"Thread for {f.__qualname__} was not running, starting again")
                        self._service_threads[f] = t = threading.Thread(target=f,name=f"{self._SERVICE_NAME}.service")
                        t.start()
                    else:
                        # self.logger.debug(f"Thread for {f.__qualname__} already running")
                        continue
                else:
                    self.logger.info(f"Creating Thread for {f.__qualname__}")
                    t = threading.Thread(target=f,name=f"{self._SERVICE_NAME}.service")
                    t.start()
                    self._service_threads[f] = t
            if self.restart:
                self._stop_threads()
                self.restart = False
            sleep(self._SERVICE_CHECK_TIME)

    def _stop_threads(self) -> bool:
        """Wait for all threads to stop
        @retval bool Threads stopped or timed out waiting
        """
        running_threads:int = len(self._service_threads)
        wait_count:int = 0
        service_threads:dict[typing.Callable,threading.Thread]
        error:bool = False
        while running_threads > 0:
            self.logger.info(f"Waiting for {str(running_threads)} thread(s) to stop")
            if wait_count > self._SERVICE_SHUTDOWN_LIMIT:
                self.logger.error("Timed out waiting for all threads to shutdown. This is likely a developer error")
                for f in self._service_threads.keys():
                    self.logger.error(f"Running Thread: {f.__qualname__}")
                    error = True
                break
            service_threads = self._service_threads.copy()
            for f,t in service_threads.items():
                if t.is_alive():
                    self.logger.debug(f"Waiting for {f.__qualname__} to finish its thread.")
                else:
                    self.logger.debug(f"Thread for {f.__qualname__} has finished.")
                    self._service_threads.pop(f)
            wait_count += 1
            running_threads = len(self._service_threads)
            sleep(1)
        return error

    def stop(self) -> None:
        """Stop Service Thread(s)
        @retval None Nothing
        """
        error:bool = self._stop_threads()
        self._pid_file.unlink()
        if error:
            sys_exit(4)

    # def service(self) -> None:
    #     """Service Thread Functionality
    #     @retval None Nothing
    #     """
    #     pass



# Copyright 2023-2023 by AccidentallyTheCable <cableninja@cableninja.net>.
# All rights reserved.
# This file is part of AccidentallyTheCables Utility Kit,
# and is released under "GPLv3". Please see the LICENSE
# file that should have been included as part of this package.
#### END COPYRIGHT BLOCK ###
import typing

def version_locator(search_version_str:str, version_list:list[str]) -> list[str]:
    """Using a Version search string, find versions which match from version_list
    @param str \c search_version_str Search String, format is `<comparator>:<version>,...`
    @param list[str] \c version_list List of Versions to search through
    @retval list[str] Versions which match the specified search string.
    """
    result:list[str] = []
    parts_versions:list[Version] = []
    parts_checks:list[str] = []
    for p in search_version_str.split(','):
        s:list[str] = p.split(':')
        parts_versions.append(Version(s[1]))
        parts_checks.append(s[0])
    for v in version_list:
        version:Version = Version(v)
        matched:bool = True
        for i in range(0,len(parts_versions)):
            check_mode:str = parts_checks[i]
            check_vers:Version = parts_versions[i]
            if check_mode == ">=":
                matched = version >= check_vers
            elif check_mode == "<=":
                matched = version <= check_vers
            elif check_mode == "==":
                matched = version == check_vers
            elif check_mode == "!=":
                matched = version != check_vers
            elif check_mode == ">":
                matched = version > check_vers
            elif check_mode == "<":
                matched = version < check_vers
            else:
                raise SyntaxError("Invalid Version Search comparator for Version")
            if not matched:
                break
        if matched:
            result.append(v)
    return result

def version_search_merge(search_a:str,search_b:str) -> str:
    """Version Search String Merger
    Combine 2 Version Search Strings (`<comparator>:<version>,...`) to form a single viable search
    @param str \c search_a Version Search String 1
    @param str \c search_b Version Search String 2
    @retval Combined Search String
    """
    versions:list[str] = []
    parts:list[str] = []
    # Create Versions and Comparators for Each Search String
    for part in search_a.split(','):
        s:list[str] = part.split(':')
        parts.append(s[0])
        versions.append(s[1])
    for part in search_b.split(','):
        s:list[str] = part.split(':')
        parts.append(s[0])
        versions.append(s[1])

    result_versions:list[str] = []
    result_output:list[str] = []

    # Create 1 query with combined searches, and locate the versions from search that match
    result_versions = version_locator(f"{search_a},{search_b}",versions)
    # Locate Comparators for each version, to build the final search
    for result_version in result_versions:
        result_comparator:str = parts[versions.index(result_version)]
        result_output.append(f"{result_comparator}:{result_version}")
    # Ensure we include version exclusions as well, if they exist
    if "!=" in parts:
        for i in range(0,len(parts)):
            if parts[i] == "!=":
                nv:str = versions[i] # Pull the exclusion version
                result_out_str:str = ','.join(result_output)
                # We may have not found any compatible versions, maybe theres only `!=` comparator entries
                if len(result_out_str) == 0:
                    result_output.append(f"!=:{nv}")
                    continue
                # Check if the exclusion version is in the current combined search result, if so, include it
                if len(version_locator(result_out_str,[nv])) > 0:
                    result_output.append(f"!=:{nv}")
    if len(result_output) == 0:
        raise ValueError(f"Unable to construct a search string which satisfies given search: '{search_a},{search_b}'")
    result:str = ','.join(result_output)
    return result


# mypy linting is disabled for this file because the unions dont play with bytes and other various things
# changes to this file should still make a best effort to be lint free.
# mypy: ignore-errors
class Version:
    """Version Breakdown Class
    A Version can be created from:
     - Semantic String ("1.0.0")
     - List of Strings or Ints of a version (["1","0","0"] or [1,0,0])
     - Tuple of Strings or Ints of a version (("1","0","0") or (1,0,0))

    Versions are comparable (>,<,>=,<=,==,!=)
    Versions are addable and subtractable (a -= b, a += b)
     - If a part goes negative, it will be set to 0

    @throws ValueError Mixed type add/subtract (str and int attempted to be added/subtracted)
    @throws TypeError Comparing Version to NoneType
    """

    parts:list[typing.Union[str,int]]

    @property
    def empty(self) -> bool:
        """Empty/None Set Version Check
        @retval bool Whether Version.parts is defined
        """
        return not hasattr(self,"parts")

    def __init__(self,version:typing.Union[
        str,
        list[
            typing.Union[str,int]
        ],
        tuple[
            typing.Union[str,int],
            ...
        ],
        None
    ]) -> None:
        if version is None:
            return
        self.parts = []
        if isinstance(version,str):
            version = version.split('.')
        elif isinstance(version,tuple):
            version = list(version)
        if version is None:
            return
        for i in range(0,len(version)):
            try:
                version[i] = int(version[i])
            except BaseException:
                version[i] = str(version[i])
            self.parts.append(version[i])

    def __str__(self) -> str:
        if not hasattr(self,"parts"):
            raise ValueError("Cannot convert empty Version to string")
        parts:list[typing.Union[str,int]] = self.parts.copy()
        out:str = str(parts.pop(0))
        for p in parts:
            out += f".{str(p)}"
        return out

    def __iadd__(self,version: typing.Union[
        str,
        list[
            typing.Union[str,int]
        ],
        tuple[
            typing.Union[str,int],
            ...
        ],
        "Version",
        None
    ]) -> "Version":
        comp:"Version" = self.__to_version(version)
        r:range = self.__define_range(comp)
        for i in r:
            if isinstance(self.parts[i],int) and isinstance(comp.parts[i],int):
                self.parts[i] += comp.parts[i]
            elif isinstance(self.parts[i],str) and isinstance(comp.parts[i],str):
                self.parts[i] = comp.parts[i]
            else:
                raise ValueError("Cannot add mixed types")
        return self

    def __add__(self,version: typing.Union[
        str,
        list[
            typing.Union[str,int]
        ],
        tuple[
            typing.Union[str,int],
            ...
        ],
        "Version",
        None
    ]) -> "Version":
        out:"Version" = Version(self.__str__())
        out += version
        return out

    def __isub__(self,version: typing.Union[
        str,
        list[
            typing.Union[str,int]
        ],
        tuple[
            typing.Union[str,int],
            ...
        ],
        "Version",
        None
    ]) -> "Version":
        comp:"Version" = self.__to_version(version)
        r:range = self.__define_range(comp)
        for i in r:
            if isinstance(self.parts[i],int) and isinstance(comp.parts[i],int):
                self.parts[i] -= comp.parts[i]
                if self.parts[i] < 0:
                    self.parts[i] = 0
                    # s:int = i - 1
                    # up:bool = True
                    # while up:
                    #     self.parts[s] -= 1
                    #     if self.parts[s] >= 0:
                    #         up = False
                    #     elif self.parts[s] < 0:
                    #         self.parts[s] = 0
                    #     s -= 1
            elif isinstance(self.parts[i],str) and isinstance(comp.parts[i],str):
                self.parts[i] = comp.parts[i]
            else:
                raise ValueError("Cannot subtract mixed types")
        return self

    def __sub__(self,version: typing.Union[
        str,
        list[
            typing.Union[str,int]
        ],
        tuple[
            typing.Union[str,int],
            ...
        ],
        "Version",
        None
    ]) -> "Version":
        out:"Version" = Version(self.__str__())
        out -= version
        return out

    def __ge__(self,version: typing.Union[
        str,
        list[
            typing.Union[str,int]
        ],
        tuple[
            typing.Union[str,int],
            ...
        ],
        "Version",
        None
    ]) -> bool:
        comp:"Version" = self.__to_version(version)
        r:range = self.__define_range(comp)
        result:bool = True
        for i in r:
            s:bytes = b''
            c:bytes = b''
            if isinstance(self.parts[i],int):
                s = bytes(self.parts[i])
            elif isinstance(self.parts[i],str):
                s = bytes(self.parts[i],"utf-8")
            if isinstance(comp.parts[i],int):
                c = bytes(comp.parts[i])
            elif isinstance(comp.parts[i],str):
                c = bytes(comp.parts[i],"utf-8")
            result = result and s >= c
            if result:
                return True
        return result

    def __le__(self,version: typing.Union[
        str,
        list[
            typing.Union[str,int]
        ],
        tuple[
            typing.Union[str,int],
            ...
        ],
        "Version",
        None
    ]) -> bool:
        comp:"Version" = self.__to_version(version)
        r:range = self.__define_range(comp)
        result:bool = True
        for i in r:
            s:bytes = b''
            c:bytes = b''
            if isinstance(self.parts[i],int):
                s = bytes(self.parts[i])
            elif isinstance(self.parts[i],str):
                s = bytes(self.parts[i],"utf-8")
            if isinstance(comp.parts[i],int):
                c = bytes(comp.parts[i])
            elif isinstance(comp.parts[i],str):
                c = bytes(comp.parts[i],"utf-8")
            result = result and s <= c
            if result:
                return True
        return result

    def __gt__(self,version: typing.Union[
        str,
        list[
            typing.Union[str,int]
        ],
        tuple[
            typing.Union[str,int],
            ...
        ],
        "Version",
        None
    ]) -> bool:
        comp:"Version" = self.__to_version(version)
        r:range = self.__define_range(comp)
        result:bool = True
        for i in r:
            s:bytes = b''
            c:bytes = b''
            if isinstance(self.parts[i],int):
                s = bytes(self.parts[i])
            elif isinstance(self.parts[i],str):
                s = bytes(self.parts[i],"utf-8")
            if isinstance(comp.parts[i],int):
                c = bytes(comp.parts[i])
            elif isinstance(comp.parts[i],str):
                c = bytes(comp.parts[i],"utf-8")
            result = result and s > c
            if result:
                return True
        return False

    def __lt__(self,version: typing.Union[
        str,
        list[
            typing.Union[str,int]
        ],
        tuple[
            typing.Union[str,int],
            ...
        ],
        "Version",
        None
    ]) -> bool:
        comp:"Version" = self.__to_version(version)
        r:range = self.__define_range(comp)
        result:bool = True
        for i in r:
            s:bytes = b''
            c:bytes = b''
            if isinstance(self.parts[i],int):
                s = bytes(self.parts[i])
            elif isinstance(self.parts[i],str):
                s = bytes(self.parts[i],"utf-8")
            if isinstance(comp.parts[i],int):
                c = bytes(comp.parts[i])
            elif isinstance(comp.parts[i],str):
                c = bytes(comp.parts[i],"utf-8")
            result = result and s < c
            if result:
                return True
        return result

    def __eq__(self,version:object) -> bool:
        comp:"Version" = self.__to_version(version)
        r:range = self.__define_range(comp)
        result:bool = True
        for i in r:
            result = result and self.parts[i] == comp.parts[i]
        return result

    def __ne__(self,version:object) -> bool:
        comp:"Version" = self.__to_version(version)
        r:range = self.__define_range(comp)
        result:bool = True
        for i in r:
            result = self.parts[i] != comp.parts[i]
            if result:
                return True
        return False

    def __to_version(self,version:object) -> "Version":
        if version is None:
            raise TypeError("Cannot Convert NoneType to Version")
        if isinstance(version,type(self)):
            return version
        return Version(version)

    def __define_range(self,version:"Version") -> range:
        if not hasattr(self,"parts") or not hasattr(version,"parts"):
            raise ValueError("Cannot compare an empty Version")
        if len(self.parts) >= len(version.parts):
            r = range(0,len(version.parts))
        else:
            r = range(0,len(self.parts))
        return r
